/*
** Module   : ControlSetEntity.java
** Abstract : server side grouped control widget parent class
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23487 Created initial version.
** 002 SIY 20060209   @24430 Major refactoring.
** 003 GES 20060803   @28371 Some cleanup and refactoring.
** 004 GES 20061103   @30943 Fixed abend due to the use of Arrays.asList()
**                           which returns an unmodifiable list.
** 005 GES 20061108   @31026 Code and doc cleanup and shift away from the character.getValue().
** 006 GES 20061128   @31452 Fixed setListItemPairs() to add in the order specified rather than
**                           reversing the order. This only affects non-sorted lists as the
**                           sorted lists would be forced into the right order later.
** 007 SIY 20061213   @31673 Maintain dynamic pairs attribute.
** 008 SIY 20101122          Inferred generics.
** 009 SIY 20101124          Fixed UnsupportedOperationException.
** 010 SVL 20101214          Added addFirst(character, character).
** 011 ECF 20110426          Repeated fix #004 for 2 more locations.
** 012 SIY 20111101          Updated imports.
** 013 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 014 OM  20130527          Added int64 support fopr entry().
** 015 CA  20131115          Changes related to properly instantiate dynamic resources.
** 016 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 017 SVL 20141102          Added isSort.
** 018 GES 20141124          Added addLast and delete methods. Enhanced addFirst. Added 
**                           list-items attr and a variant of entry().
** 019 IAS 20150123          Implemented DELETE() method
** 020 IAS 20150202          Implemented ADD_LAST() method, fixed ADD-FIRST()
** 021 IAS 20150225          Added support for the DELIMITER attribute,
*                            fixed support for the LIST-ITEMS and LIST-ITEM-PAIRS attributes
*/

package com.goldencode.p2j.ui;

import java.util.*;
import java.util.regex.*;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.client.format.*;

/**
 * Server side parent class for widgets that group other controls.
 */
public abstract class ControlSetEntity<T extends ControlSetConfig>
extends ControlEntity<T> 
implements CommonListWidget<T>
{
   /* Predicate to check ControlSetItem label */
   protected static final BiPredicate<String, ControlSetItem> COMPARE_LABEL = 
         new BiPredicate<String, ControlSetItem>()
   {
      @Override
      public boolean test(String label, ControlSetItem item)
      {
         return label.equals(item.getLabel().toStringMessage());
      }
   };

   /* Predicate to check ControlSetItem label case-insensitive*/
   protected static final BiPredicate<String, ControlSetItem> COMPARE_LABEL_IGNORE_CASE = 
         new BiPredicate<String, ControlSetItem>()
   {
      @Override
      public boolean test(String label, ControlSetItem item)
      {
         return label.equalsIgnoreCase(item.getLabel().toStringMessage());
      }
   };

   /* Function to check ControlSetItem value */
   protected final BiPredicate<String, ControlSetItem> COMPARE_VALUE_IGNORE_CASE = 
         new BiPredicate<String, ControlSetItem>()
   {
      @Override
      public boolean test(String value, ControlSetItem item)
      {
         return truncateValue(value).equalsIgnoreCase(item.getValue().toStringMessage());
      }
   };
   
   /** trailing spaces */
   protected static final Pattern RTRIM = Pattern.compile("\\s+$"); 

   /** leading spaces */
   protected static final Pattern LTRIM = Pattern.compile("^\\s+");

   /**
    * Construct an instance using the given configuration.
    * 
    * @param   dynamic
    *          Flag indicating if this is a static or dynamic resource.
    * @param   csc
    *          Source configuration.
    */
   public ControlSetEntity(boolean dynamic, T csc)
   {
      super(dynamic, csc);
   }
   
   /**
    * Implements the ENTRY() widget method.  This method returns items from
    * the set by 1-based index.
    *
    * @param    index
    *           The index of an item.
    *
    * @return   The list item indexed or <code>unknown value</code>.
    */
   @Override
   public character entry(int64 index)
   {
      if (index == null || index.isUnknown())
         return new character();
      
      return entry(index.longValue());
   }

   /**
    * Implements the ENTRY() widget method.  This method returns items from
    * the set by 1-based index.
    *
    * @param    index
    *           The index of an item.
    *
    * @return   The list item indexed or <code>unknown value</code>.
    */
   @Override
   public character entry(long index)
   {
      ControlSetItem[] items = config.items;
      
      if (index < 1 || index > items.length)
         return new character();

      return items[(int)index - 1].getLabel();
   }

   /**
    * Implements the LOOKUP() widget method.  This method returns the 1-based
    * item's index in the set.
    *
    * @param    text
    *           The text of an item.
    *
    * @return   The item's index or <code>unknown value</code>.
    */
   @Override
   public integer lookup(character text)
   {
      if (text.isUnknown())
         return new integer();

      ControlSetItem[] items = config.items;
      
      for (int i = 0; i < items.length; i ++)
      {
         if (items[i].getLabel().compareTo(text) == 0)
             return new integer(i + 1);
      }

      // if list-string is not in the list, LOOKUP returns 0
      return new integer(0);
   }
   
   /**
    * Implements the LOOKUP() widget method.  This method returns the 1-based
    * item's index in the set.
    *
    * @param    text
    *           The text of an item.
    *
    * @return   The item's index or <code>unknown value</code>.
    */
   @Override
   public integer lookup(String text)
   {
      return lookup(new character(text));
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *           
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addFirst(character itemList)
   {
      if (itemList == null)
      {
         itemList = new character(); 
      }

      return addFirst(itemList.toStringMessage());
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addFirst(String itemList)
   {
      return add(itemList, new ItemsListExpander()
      {
         @Override
         public String method()
         {
            return "ADD-FIRST";
         }

         @Override
         public void addItems(String[] elem, ControlSetItem[] items, List<ControlSetItem> list)
         {
            for (String s : elem)
            {
               if (!ignore(s))
               {
                  list.add(controlSetItem(s));
               }
            }
            list.addAll(Arrays.asList(items));
         }
      });
   }

   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   public logical addFirst(character label, BaseDataType value)
   {
      value = checkValue("ADD-FIRST", label, value);
 
      if (value == null)
      {
         return new logical(false);
      }
      
      return add(label, value, new ItemsListByOneExpander()
      {
         @Override
         public ControlSetItem[] addItem(ControlSetItem[] items, character label, 
               BaseDataType value)
         {
            ControlSetItem[] nitems = new ControlSetItem[items.length + 1];
            nitems[0] = controlSetItem(label, truncateValue(value));
            System.arraycopy(items, 0, nitems, 1, items.length);
            return nitems;
         }
      });
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   public logical addFirst(String label, BaseDataType value)
   {
      return addFirst(new character(label), value);
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   public logical addFirst(String label, String value)
   {
      return addFirst(new character(label), new character(value));
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   public logical addFirst(character label, String value)
   {
      return addFirst(label, new character(value));
   }
   
   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character itemList)
   {
      if (itemList == null)
      {
         itemList = new character(); 
      }
      
      return addLast(itemList.toStringMessage());
   }
   
   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(String itemList)
   {
      return add(itemList, new ItemsListExpander()
      {
         @Override
         public String method()
         {
            return "ADD-LAST";
         }

         @Override
         public void addItems(String[] elem, ControlSetItem[] items, List<ControlSetItem> list)
         {
            list.addAll(Arrays.asList(items));
            for (String s : elem)
            {
               if (!ignore(s))
               {
                  list.add(controlSetItem(s));
               }
            }
         }
      });
   }
   
   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character label, BaseDataType value)
   {
      value = checkValue("ADD-LAST", label, value);

      if (value == null)
      {
         return new logical(false);
      }

      return add(label, value, new ItemsListByOneExpander()
      {
         @Override
         public ControlSetItem[] addItem(ControlSetItem[] items, character label, 
               BaseDataType value)
         {
            ControlSetItem[] nitems = Arrays.copyOf(items, items.length + 1);
            nitems[items.length] = controlSetItem(label, value);
            return nitems;
         }
      });
   }

   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(String label, BaseDataType value)
   {
      return addLast(new character(label), value);
   }

   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(character label, String value)
   {
      return addLast(label, new character(value));
   }

   /**
    * Implements the ADD-LAST() widget method, which adds a list of labels
    * to the end of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addLast(String label, String value)
   {
      return addLast(new character(label), new character(value));
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the given index.
    *
    * @param    idx
    *           The index of the item to delete.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(int64 idx)
   {
      if (idx == null || idx.isUnknown())
         return new logical();

      return delete(idx.longValue());
   }

   /**
    * Implements the DELETE() widget method, which removes the list item at the given index.
    *
    * @param    idx
    *           The index of the item to delete.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical delete(long idx)
   {
      ControlSetItem[] items = config.items;

      if (idx < 0 || idx > items.length)
         return new logical(false);
      if (idx == 0)
      {
         ErrorManager.recordOrShowError(4065, String.format(
               "The DELETE attribute on the %s %s has invalid arguments", type(),
               frame.getName(this)), false);
         return new logical(false);
      }
      idx--;
      ControlSetItem[] nitems = new ControlSetItem[items.length - 1];
      int nidx = 0;
      for (int i = 0; i < items.length; i++)
      {
         if (i == (int) idx)
            continue;
         nitems[nidx++] = items[i];
      }
      setItems(nitems);
      return new logical(true);
   }
   
   /**
    * Store array of items into config.
    * 
    * @param    items
    *           Array of new items.
    * @param    pairs
    *           true if LIST_ITEM_PAIRS, RADIO_BUTTONS is set
    *           false if LIST_ITEMs IS set
    */
   public void setItems(ControlSetItem[] items)
   {
      setItems(items, true);
   }

   /**
    * Store array of items into config.
    * 
    * @param    items
    *           Array of new items.
    * @param    pairs
    *           true if LIST_ITEM_PAIRS, RADIO_BUTTONS is set
    *           false if LIST_ITEMs IS set
    */
   @Override
   public void setItems(ControlSetItem[] items, boolean pairs)
   {
      if (config.items != null && items == null)
         return;
      items = truncateValues(items);
      for (ControlSetItem item: items)
      {
         if (item.getId() == 0)
         {
            item.setId(config.itemId++);
         }
      }

      config.items = items;
      config.pairs = pairs;
      sortItems();
      
      pushScreenDefinition();
   }
 
   /**
    * Set SORT attribute.
    * 
    * @param    sort
    *           New value for the attribute.
    */
   public void setSort(logical sort)
   {
      setSort(sort.booleanValue());
   }
   
   /**
    * Set SORT attribute.
    * 
    * @param    sort
    *           New value for the attribute.
    */
   public void setSort(boolean sort)
   {
      config.sort = sort;
      sortItems();
      pushScreenDefinition();
   }

   /**
    * Get SORT attribute.
    *
    * @return  value for the attribute.
    */
   public logical isSort()
   {
      return new logical(config.sort);
   }
   
   /**
    * Sets the LIST-ITEMS writeable attribute.
    * 
    * @param    list
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.  These
    *           items will be set as the full list of items in this widget.
    */
   @Override
   public void setListItems(character list)
   {
      if (config.pairs)
      {
         useListItemPairs(false);
         return;
      }

      if (list == null || list.isUnknown())
      {
         config.items = null;
         pushScreenDefinition();
         return;
     }

      setItems(new ControlSetItem[0]);
      
      addFirst(list);
   }
   
   /**
    * Sets the LIST-ITEMS writeable attribute.
    * 
    * @param    list
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.  These
    *           items will be set as the full list of items in this widget.
    */
   @Override
   public void setListItems(String list)
   {
      setListItems(new character(list));
   }
   
   /**
    * Common implementation list item pair attributes (LIST-ITEM-PAIRS and
    * RADIO-BUTTONS) attribute. The parameter contains string representation
    * of all labels and values separated by a delimiter (comma is the default
    * delimiter).
    * <p>
    * Example: <code>Label1,value1,Label2,value2,...,LabelN,valueN</code>
    * 
    * @param    list
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.  These
    *           items will be set as the full list of items in this widget.
    */
   @Override
   public void setListItemPairs(String list)
   {
      setListItemPairs(new character(list));
   }
   
   /**
    * Common implementation list item pair attributes (LIST-ITEM-PAIRS and
    * RADIO-BUTTONS) attribute. The parameter contains string representation
    * of all labels and values separated by a delimiter (comma is the default
    * delimiter).
    * <p>
    * Example: <code>Label1,value1,Label2,value2,...,LabelN,valueN</code>
    * 
    * @param    list
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.  These
    *           items will be set as the full list of items in this widget.
    */
   @Override
   public void setListItemPairs(character list)
   {
      if (!config.pairs)
      {
         useListItems(false);
         return;
      }

      if (list == null || list.isUnknown())
      {
         config.items = null;
         pushScreenDefinition();
         return;
     }

      String itemList = list.toStringMessage();
      char[] delim = new char[1];
      delim[0] = config.delimiter;
      String d = new String(delim);
      String type = config.dataType;
      String[] elem = itemList.split(d);
      if (elem.length % 2 != 0)
      {
         ErrorManager.recordOrShowError(7452, 
               String.format(
                  "LIST-ITEM-PAIRS must contain an even number of entries to specify label/value pairs for %s %s.", 
                  type(), frame.getName(this)
               ), 
               false);
         ErrorManager.recordOrShowError(4058, 
               String.format(
                  "**Attribute LIST-ITEM-PAIRS for the %s %s has an invalid value of %s.", 
                  type(), frame.getName(this), list.toStringMessage()
               ), 
               false);
         
      }
      ControlSetItem[] items = config.items;

      if (items == null)
         items = new ControlSetItem[0];

      List<ControlSetItem> newlist = asList(items);
      
      for (int i = 0; i < elem.length; i += 2)
      {
         character label = new character(elem[i]);
         BaseDataType value = DisplayFormat.instanceOfType(type, truncateValue(elem[i + 1]));
         if (invalid(label, value))
         {
            continue;
         }
         newlist.add(new ControlSetItem(config.itemId++, 
               label, 
               DisplayFormat.instanceOfType(type, truncateValue(elem[i + 1]))));
      }
      
      items = newlist.toArray(items);
      
      config.pairs = true;
      setItems(items);
      pushScreenDefinition();
   }
   
   /**
    * Gets the LIST-ITEMS writeable attribute.
    * 
    * @return   The current value of LIST-ITEMS attribute.
    */
   @Override
   public character getListItems()
   {
      if (config.pairs)
      {
         useListItemPairs(true);
         return new character();
      }

      ControlSetItem[] items = config.items;

      if (items == null)
         items = new ControlSetItem[0];
      
      StringBuffer sb = new StringBuffer();
      
      for (int i = 0; i < items.length; i++)
      {
         sb.append(items[i].getLabel().toStringMessage());
         
         if (i < items.length - 1)
            sb.append(config.delimiter);
      }
      
      return new character(sb.toString());
   }

   /**
    * Gets the LIST-ITEM-PAIRS writeable attribute.
    *
    * @return   The current value of LIST-ITEMS attribute.
    */
   @Override
   public character getListItemPairs()
   {
      if (!config.pairs)
      {
         useListItems(true);
         return new character();
      }

      ControlSetItem[] items = config.items;

      if (items == null)
         items = new ControlSetItem[0];
      
      StringBuffer sb = new StringBuffer();
      char delimiter = config.delimiter;
      
      for (int i = 0; i < items.length; i++)
      {
         sb.append(items[i].getLabel().toStringMessage());
         sb.append(delimiter);
         sb.append(items[i].getValue().toStringMessage());
         if (i < items.length - 1)
            sb.append(delimiter);
      }
      
      return new character(sb.toString());
   }
   
   /**
    * Tell widget that data format should be honored.
    * 
    * @param    b
    *           If <code>true</code>, the format string will be honored.
    */
   public void setHonorFormat(boolean b)
   {
      config.honorFormat = b;
      pushScreenDefinition();
   }

   /**
    * Gets the DELIMITER writeable attribute.
    *
    * @return   The current value of DELIMITER attribute.
    */
   @Override
   public character getDelimiter() 
   {
      return new character(config.delimiter);
   }

   /**
    * Sets the DELIMITER writeable attribute.
    *
    * @param value
    *       The new value of DELIMITER attribute.
    */
   @Override
   public void setDelimiter(character value) 
   {
      if (value == null || value.isUnknown())
      {
         ErrorManager.recordOrShowError(4083, String.format(
               "Unable to assign UNKNOWN value to attribute DELIMITER on %s %s.", type(),
               frame.getName(this)), false);
         return;
      }
      setDelimiter(value.toStringMessage());
   }

   
   /**
    * Sets the DELIMITER writeable attribute.
    *
    * @param value
    *       The new value of DELIMITER attribute.
    */
   @Override
   public void setDelimiter(String value) 
   {
      if (value.length() != 1 || value.charAt(0)< 1 || value.charAt(0) > 127 )
      {
         ErrorManager.recordOrShowError(4058, String.format(
               "Attribute DELIMITER for the %s %s has an invalid value of %s.", type(),
               frame.getName(this), value), false);
         return;
      }
      config.delimiter = value.charAt(0);
   }
  /**
    * Helper method: append/prepend item to config.items
    * @param label 
    *          label of item to add 
    * @param value 
    *          value of item to add 
    * @param expander
    *          expander which will perform append/prepend
    * @return   <code>true</code> if the method succeeds.
    */
   protected logical add(character label, BaseDataType value, ItemsListByOneExpander expander)
   {
      if (!config.pairs) 
      {
         ErrorManager.recordOrShowError(8950, String.format(
               "Ignoring extra argument(s) of ADD-LAST for %s %s that is of LIST-ITEMS type", 
               type(), frame.getName(this)), false);
         return addLast(label);
      }
      ControlSetItem[] items = config.items;
      if (items == null)
         items = new ControlSetItem[0];
      setItems(expander.addItem(items, label, value));
      return new logical(true);
   }

   /**
    * Helper method: append/prepend items to config.items
    * @param itemList
    *          items to add 
    * @param expander
    *          expander which will perform append/prepend
    * @return   <code>true</code> if the method succeeds.
    */
   protected logical add(String itemList, ItemsListExpander expander)
   {
      if (itemList == null) 
      {
         itemList = new character().toStringMessage();
      }
      
      if (config.pairs) 
      {
         ErrorManager.recordOrShowError(4064, String.format(
               "Missing argument when setting attribute %s for %s %s", type(),
               expander.method(), frame.getName(this)), false);
         ErrorManager.recordOrShowError(4065, String.format(
               "The ADD-LAST attribute on the %s %s has invalid arguments", type(),
               frame.getName(this)), false);
         return new logical(false);
      }
      
      String[] elem = itemList.split(String.valueOf(config.delimiter));
      ControlSetItem[] items = config.items;
      if (items == null)
         items = new ControlSetItem[0];
      
      List<ControlSetItem> list = new ArrayList<>(items.length + elem.length);
      expander.addItems(elem, items, list);
      
      setItems(list.toArray(new ControlSetItem[list.size()]));
      return new logical(true);
   }

   /**
    * Create new ControlSetItem
    * 
    * @param label
    *       item' label
    * @param value
    *       item's value
    *       
    * @return 
    *    new instance of the ControlSetItem
    */
   protected ControlSetItem controlSetItem(character label, BaseDataType value) 
   {
      return new ControlSetItem(config.itemId++, 
            label.isUnknown() ? new character(" ") : (character)label.duplicate(), 
            value.duplicate()); 
   }

   /**
    * Helper method for the DELETE() widget method, which removes the list item
    * or items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * @param checkItem
    *           The predicate for selection items to delete.
    * @param checkAll
    *           return TRUE only if all items where found
    * 
    * @return <code>true</code> if the method succeeds.
    */
   protected logical delete(character itemList, 
         BiPredicate<String, ControlSetItem> checkItem, boolean checkAll)
   {
      if (itemList == null ) 
      {
         itemList = new character();
      }
      return delete(itemList.toStringMessage(), checkItem, checkAll);
   }

   /**
    * Helper method for the DELETE() widget method, which removes the list item
    * or items specified.
    * 
    * @param itemList
    *           The item or items to delete.
    * @param checkItem
    *           The predicate for selection items to delete.
    * @param checkAll
    *           return TRUE only if all items where found
    * 
    * @return <code>true</code> if the method succeeds.
    */
   protected logical delete(String itemList, 
         BiPredicate<String, ControlSetItem> checkItem, boolean checkAll)
   {
      if (itemList == null) 
      {
         itemList = "?";
      }
      return delete(
            new ArrayList<String>(Arrays.asList(itemList.split(new String(
                  new char[] { config.delimiter })))), checkItem, checkAll);
   }

   /**
    * Helper method for the DELETE() widget method, which removes the list item
    * or items specified.
    * 
    * @param itemList
    *           The List of items to delete.
    * @param checkAll
    *           return TRUE only if all items where found
    * 
    * @return <code>true</code> if the method succeeds.
    */
   protected logical delete(List<String> itemList, 
         BiPredicate<String, ControlSetItem> checkValue, boolean checkAll)
   {
      boolean foundLast = false;
      ControlSetItem[] items = config.items;
      List<ControlSetItem> nitems = new ArrayList<>(items.length);
      for (ControlSetItem item : items)
      {
         int ip = -1;
         for (int i = 0; i < itemList.size(); i++)
         {
            if (checkValue.test(itemList.get(i),  item))
            {
               ip = i;
               break;
            }
         }
         if (ip < 0)
         {
            nitems.add(item);
            continue;
         }
         if (ip == itemList.size() - 1)
         {
            foundLast = true;
         }
         itemList.remove(ip);
      }
      setItems(nitems.toArray(new ControlSetItem[nitems.size()]));
      return new logical(checkAll ? itemList.isEmpty() : foundLast);
   }

   /**
    * Check if a label only item is acceptable
    * 
    * @param attr
    *       the attribute in question
    *        
    * @return <code>true</code> if it is OK.
    */
   protected boolean checkIfNotPairs(String attr)
   {
      if (config.pairs)
      {
         ErrorManager.recordOrShowError(4064, 
               String.format(
                  "**Missing argument when setting attribute %s for %s %s.", 
                  attr, type(), frame.getName(this)
               ), 
               false);
         ErrorManager.recordOrShowError(4065, 
               String.format(
                  "**The %s attribute on the %s %s has invalid arguments.", 
                  attr, type(), frame.getName(this)
               ), 
               false);
         return false;
      }
      return true;
   }

   /**
    * Check if a label/value item is acceptable
    * 
    * @param attr
    *       the attribute in question 
    * @param label
    *       item's label
    * @param value
    *       item's value
    *       
    * @return <code>true</code> if it is OK.
    */
   protected BaseDataType checkValue(String attr, character label, BaseDataType value)
   {
      if (!config.pairs)
      {
         ErrorManager.recordOrShowError(8590, 
               String.format(
                  "Ignoring extra argument(s) of %s for %s %s that is of LIST-ITEMS type.", 
                  attr, type(), frame.getName(this)
               ), 
               false);
         return truncateValue(label);
      } 
      
      /*
       * 4GL makes difference between UNKNOWN literal and the UNKNOWN value
       * the first one is not acceptable as value while the second one is OK
       */
      if (value instanceof unknown)
      {
         ErrorManager.recordOrShowError(4067, 
            String.format(
               "**The second argument for %s on the %s %s should not be of type UNKNOWN.", 
               attr, type(), frame.getName(this)
            ), 
            false);
         return null;
      }

      if (invalid(label, value))
      {
         // The message looks weird but this is how 4GL works
         // Which is more weird sometimes an emitted warning is 4059
         // with 'FALSE' instead of '0' at the end. 
         // It seems that it happens only once - at the first attempt 
         // to add an invalid (label, value) pair
         ErrorManager.recordOrShowError(4057, 
               String.format(
                  "**Attribute %s for the %s %s has an invalid value of 0.", 
                  attr, type(), frame.getName(this)
               ), 
               false);
      }
      return truncateValue(value);
   }

   /**
    * Remove trailing whitespaces
    * 
    * @param value
    *    the string to be truncated
    *    
    * @return the truncated argument
    */
   protected String truncateValue(String value)
   {
      return value;
   }

   /**
    * Remove trailing whitespaces if applicable
    * 
    * @param value
    *    the BaseDataType to be truncated
    *    
    * @return
    *    the truncated argument if the argument is of type character
    *    (nothing is required for non-character types)
    */
   protected BaseDataType truncateValue(BaseDataType value)
   {
      return value;
   }

   /**
    * Truncate values of the provided items and remove invalid ones.
    * 
    * @param items
    *    items to be processed
    *    
    * @return
    *    valid items with truncated values
    */
   protected ControlSetItem[] truncateValues(ControlSetItem[] items)
   {
      return items;
   }

   /**
    * Checks if a (label,value) pair is 
    *       not acceptable for the widget
    *       (("?","?") for the COMBO-BOX)
    *       
    * @param label 
    *       label to be checked
    * @param value 
    *       value to be checked
    *    
    * @return <code>true</code> if the value should be ignored
    */
   protected boolean invalid(character label, BaseDataType value)
   {
      return ignore(label) && ignore(value);
   }

   /**
    * Checks if a value should be ignored ("?" for the COMBO-BOX)
    *  
    * @param value
    *          value to be checked
    *    
    * @return <Code>true</code> if the value should be ignored
    */
   protected boolean ignore(String value)
   {
      return false;
   }

   /**
    * 
    * Checks if a value should be ignored (UNKNOWN for COMBO-BOX)
    *  
    * @param 
    *    value - value to be checked
    *    
    * @return <code>true</code> if the value should be ignored
    */
   protected boolean ignore(BaseDataType value)
   {
      return false;
   }

   /**
    * Emits warning about incorrect access to the LIST-ITEMS attribute
    * 
    * @param getter
    *       specifies if it was a read access
    */
   protected void useListItemPairs(boolean getter)
   {
      ErrorManager.recordOrShowError(7454, 
            String.format(
            "For %s %s specified with Label/Value pairs, " +
            "use LIST-ITEM-PAIRS instead of LIST-ITEMS.", 
            type(), frame.getName(this)), false);
        
      ErrorManager.recordOrShowError(4052, 
            String.format(
            "**LIST-ITEMS is not a queryable attribute for %s %s.",
            getter ? "queryable" : "setable",
            type(), frame.getName(this)), false);
      
   }
   /**
    * Emits warning about incorrect access to the LIST-ITEM-PAIRS attribute
    * 
    * @param getter
    *          specifies if it was a read access
    */
   protected void useListItems(boolean getter)
   {
      ErrorManager.recordOrShowError(7453, 
            String.format(
            "For %s %s specified with only labels, use LIST-ITEMS instead of LIST-ITEM-PAIRS.", 
            type(), frame.getName(this)), false);
     
      ErrorManager.recordOrShowError(4052, 
            String.format(
            "**LIST-ITEM-PAIRS is not a %s attribute for %s %s.", 
            getter ? "queryable" : "setable",
            type(), frame.getName(this)), false);
   }
   
   /**
    * Create new ControlSetItem with value == lable 
    * 
    * @param s
    *       label and value of the item
    *
    * @return new instance of the ControlSetItem 
    */
   private ControlSetItem controlSetItem(String s) 
   {
      return controlSetItem(new character("?".equals(s.trim()) ? "" : s), 
            new character(truncateValue(s))); 
   }

   /**
    * Convert the given array to a list that is modifiable.  Note that the
    * <code>Arrays.asList()</code> method does NOT return a modifiable list.
    *
    * @param    items
    *           The array to be converted.
    *
    * @return   The modifiable list in the same order as the original array.
    */
   private List<ControlSetItem> asList(ControlSetItem[] items)
   {
      ArrayList<ControlSetItem> list = new ArrayList<ControlSetItem>();
      
      for (ControlSetItem item : items)
      {
         list.add(item);
      }
      
      return list;
   }
   
   /**
    * Sort items according to current sort flag. 
    */
   private void sortItems()
   {
      if (!config.sort || config.items == null)
         return;
      
      Arrays.sort(config.items, new ItemComparator());
   }

   /**
    * Local comparator for two Frame instances.
    */
   private static class ItemComparator
   implements Comparator<ControlSetItem>
   {
      /**
       * Compare two instances of <code>ControlSetItem</code> for label.
       * 
       * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
       */
      @Override
      public int compare(ControlSetItem o1, ControlSetItem o2)
      {
         return o1.getLabel().compareTo(o2.getLabel());
      }
   }

   /**
    * Helper interface (is standard in Java 8.
    */
   private static interface BiPredicate<T,U> 
   {
      public boolean test(T t, U u);
   }

   /** 
    * Helper interface for appending/prepending a new single item to a list 
    */
   private static interface ItemsListByOneExpander 
   {
      /**
       * Expand ControlSetItem array by one item
       *   
       * @param items
       *       array to expand
       * @param label
       *       label of the new items 
       * @param value
       *       value of the new item
       *
       * @return
       *       expanded array
       */
      public ControlSetItem[] addItem(ControlSetItem[] items, 
            character label, BaseDataType value);
   }

   /** 
    * Helper interface for appending/prepending multiple items to a list 
    */
   private static interface ItemsListExpander 
   {
      /**
       * Expand ControlSetItem array by multiple items
       *   
       * @param elem
       *        labels of new items
       * @param items
       *       array to expand
       * @param list
       *       list for holding the result 
       */
      public void addItems(String[] elem, ControlSetItem[] items, List<ControlSetItem> list);
      
      /**
       * Get name of the method for adding items (ADD-FIRST/ADD-LAST)
       * 
       * @return
       *    the method name
       */
      public String method();
   }
}
